import base64
import requests
import json
import time

from . import config
from .cachehandler import CacheHandler

from .constants.reasons import *
from .constants.transporters import *

from .endpoints.orders import OrderMethods
from .endpoints.shippinglabels import ShippingLabelMethods


class BolAPI:

    def __init__(self, clientId, clientSecret, demo=False):

        self.clientId = clientId
        self.clientSecret = clientSecret
        self.demo = demo
        self.headers = {
            'Accept' : 'application/vnd.retailer.v5+json',
            'Content-Type' : 'application/vnd.retailer.v5+json',
        }

        self.baseUrl = config.DEMO_URL if demo else config.BASE_URL
        self.cacheHandler = CacheHandler()

        self.rateLimitRemaining = None
        self.rateLimitReset = None

        self.orders = OrderMethods(self)
        self.shippingLabels = ShippingLabelMethods(self)

        self.reasons = Reasons()
        self.transporters = Transporters()
    
    def setTokenHeader(self, token):
        bearerStr = 'Bearer {token}'.format(token=token)
        self.headers.update({'Authorization' : bearerStr})

    def checkHeaderTokens(self):

        # If no authorization header is found, we need to include the token
        if 'Authorization' not in self.headers:
            
            # Check if we have a token stored in cache, if not, acquire one
            # If we do, set it in the header
            cachedAccessToken = self.cacheHandler.getCache(self.clientId)
            if cachedAccessToken is None:
                self.acquireAccessToken()
            else:
                self.setTokenHeader(cachedAccessToken)

    def acquireAccessToken(self):

        credentialsString = '{clientId}:{clientSecret}'.format(clientId=self.clientId, clientSecret=self.clientSecret)
        encodedBytes = base64.b64encode(credentialsString.encode('utf-8'))
        encodedString = str(encodedBytes, 'utf-8')

        credentialsHeader = {
            'Accept' : 'application/json',
            'Authorization' : 'Basic {encodedString}'.format(encodedString=encodedString)
        }

        req = requests.post(config.AUTH_URL, data=None, headers=credentialsHeader)
      
        status = req.status_code
        response = req.json()

        if status == 200:
            accessToken = response['access_token']
            self.cacheHandler.setCache(self.clientId, accessToken)
            self.setTokenHeader(accessToken)

            return accessToken

    def checkRateLimits(self, response):
        headers = response.headers

        if 'X-RateLimit-Remaining' in headers:
            self.rateLimitRemaining = int(headers['X-RateLimit-Remaining'])
        
        if 'X-RateLimit-Reset' in headers:
            self.rateLimitReset = int(headers['X-RateLimit-Reset'])


        # If we only have one request or less remaining, delay until the limit is reset
        if self.rateLimitRemaining <= 1:
            currentMil = time.time() * 1000
            waitingMil = self.rateLimitReset - currentMil

            time.sleep(waitingMil/1000)

    def doRequest(self, method, url, data=None, headers=None):

        if headers:
            mergedHeaders = self.headers
            mergedHeaders.update(headers)
            headers = mergedHeaders
        else: headers = self.headers

        reqUrl = '{base}/{url}'.format(base=self.baseUrl, url=url)

        if method == 'GET':
            response = requests.get(reqUrl, params=data, headers=headers)
        elif method == 'POST':
            response = requests.post(reqUrl, data=json.dumps(data), headers=headers)
        elif method == 'PUT':
            response = requests.put(reqUrl, data=json.dumps(data), headers=headers)
        
        return response


    def request(self, method, url, data=None, headers=None):
        
        # Check the headers for appropriate tokens before we make a request
        self.checkHeaderTokens()

        # Make the request
        response = self.doRequest(method, url, data, headers)
        
        # Check if request has been carried out
        # If request has been blocked due to token expiration, get a new one and redo the request
        if response.status_code == 401:
            self.acquireAccessToken()

            response = self.doRequest(method, url, data, headers)

        # Check the rate remaining, delay if necessary
        self.checkRateLimits(response)

        if 'json' in response.headers['Content-Type']:
            respContent = response.json()
        elif 'pdf' in response.headers['Content-Type']:
            respContent = response.content
        
        return response.status_code, response.headers, respContent
    
    def get(self, url, data=None, headers=None):
        status, headers, response = self.request('GET', url, data, headers)
        return status, headers, response
    
    def post(self, url, data=None, headers=None):
        status, headers, response = self.request('POST', url, data, headers)
        return status, headers, response
    
    def put(self, url, data=None, headers=None):
        status, headers, response = self.request('PUT', url, data, headers)
        return status, headers, response


class Reasons:
    OUT_OF_STOCK = "OUT_OF_STOCK"
    REQUESTED_BY_CUSTOMER = "REQUESTED_BY_CUSTOMER"
    BAD_CONDITION = "BAD_CONDITION"
    HIGHER_SHIPCOST = "HIGHER_SHIPCOST"
    INCORRECT_PRICE = "INCORRECT_PRICE"
    NOT_AVAIL_IN_TIME = "NOT_AVAIL_IN_TIME"
    NO_BOL_GUARANTEE = "NO_BOL_GUARANTEE"
    ORDERED_TWICE = "ORDERED_TWICE"
    RETAIN_ITEM = "RETAIN_ITEM"
    TECH_ISSUE = "TECH_ISSUE"
    UNFINDABLE_ITEM = "UNFINDABLE_ITEM"
    OTHER = "OTHER"


from .base import APIEndpoint

from bol.models.orders import OrderList, Order
from bol.models.processes import ProcessStatus

class OrderMethods(APIEndpoint):

    def __init__(self, api):
        super(OrderMethods, self).__init__(api, "orders")

    def list(self, page=1, method='FBR', status='ALL'):

        data = { 'page' : page, 'method' : method, 'status' : status }
        url = self.endpoint

        status, headers, respJson = self.api.get(url, data)
        print(respJson)
        if status == 400: return OrderList().parseError(respJson)

        return OrderList().parse(respJson)

    def get(self, id):

        url = '{endpoint}/{id}'.format(endpoint=self.endpoint, id=id)
        data = None

        status, headers, respJson = self.api.get(url, data)
        if status == 404: return Order().parseError(respJson)

        return Order().parse(respJson)
    
    def cancelItem(self, item, reason):
        data = { 'orderItems' : [{ 'orderItemId' : item.orderItemId, 'reasonCode' : reason }]}

        url = '{endpoint}/cancellation'.format(endpoint=self.endpoint)

        status, headers, respJson = self.api.put(url, data)
        if status == 400: return ProcessStatus().parseError(respJson)
        
        return ProcessStatus().parse(respJson)
    
    def cancel(self, order, reason):

        processStatuses = []

        for item in order.orderItems:
            processStatus = self.cancelItem(item, reason)
            processStatuses.append(processStatus)

        return processStatuses
    
    def shipItem(self, item, shipmentReference=None, shippingLabelId=None, transporterCode=None, trackAndTrace=None):
        data = { 'orderItems' : [{ 'orderItemId' : item.orderItemId }] }

        if shipmentReference: data['shipmentReference'] = shipmentReference
        if shippingLabelId: data['shippingLabelId'] = shippingLabelId
        if transporterCode:
            data['transport'] = { 'transporterCode' : transporterCode }
            if trackAndTrace: data['transport']['trackAndTrace'] = trackAndTrace

        url = '{endpoint}/shipment'.format(endpoint=self.endpoint)
        
        status, headers, respJson = self.api.put(url, data)
        if status == 400: return ProcessStatus().parseError(respJson)

        return ProcessStatus().parse(respJson)

    def ship(self, order, shipmentReference=None, shippingLabelId=None, transporterCode=None, trackAndTrace=None):

        processStatuses = []

        for item in order.orderItems:
            processStatus = self.shipItem(item, shipmentReference, shippingLabelId, transporterCode, trackAndTrace)
            processStatuses.append(processStatus)

        return processStatuses


from .base import ObjectListModel, BaseModel

class ProcessStatus(ObjectListModel):

    STATUS_PENDING = "PENDING"
    STATUS_SUCCESS = "SUCCESS"
    STATUS_FAILURE = "FAILURE"
    STATUS_TIMEOUT = "TIMEOUT"

    def __init__(self, 
        processStatusId=None,
        entityId=None,
        eventType=None,
        description=None,
        status=None,
        errorMessage=None,
        createTimestamp=None,
    ):

        super(ProcessStatus, self).__init__(list=[])

        self.processStatusId = processStatusId
        self.entityId = entityId
        self.eventType = eventType
        self.description = description
        self.status = status
        self.errorMessage = errorMessage
        self.createTimestamp = createTimestamp
        
        self.hasError = False
        self.error = None

    @property
    def links(self):
        return self.list
    
    def parse(self, json):
        super(ProcessStatus, self).parse(json)
        
        if 'links' in json:
            for link in json['links']:

                link = ProcessLink().parse(link)
                self.addToList(link)
        
        return self


class ProcessLink(BaseModel):

    def __init__(self,
        rel=None,
        href=None,
        method=None
    ):

        self.rel = rel
        self.href = href
        self.method = method
        
        self.hasError = False
        self.error = None